from .components import CrudComponent, ServiceEntity, KongError, KongEntity


class PluginMixin:

    def wrap(self, data):
        return Plugin.factory(self.cli, data)

    @property
    def url(self) -> str:
        return '%s/%s' % (self.cli.url, self.name)

    def get_list(self, **params):
        url = '%s/%s' % (self.root.url, self.name)
        return self.execute(url, params=params, wrap=self.wrap_list)

    async def apply_json(self, data):
        if not isinstance(data, list):
            data = [data]
        plugins = await self.get_list()
        plugins = dict(((p['name'], p) for p in plugins))
        result = []
        for entry in data:
            name = entry.pop('name', None)
            if not name:
                raise KongError('Plugin name not specified')
            if name in plugins:
                plugin = plugins.pop(name)
                plugin = await self.update(
                    plugin.id, name=name, **entry)
            else:
                plugin = await self.create(name=name, **entry)

            result.append(plugin.data)
        for entry in plugins.values():
            await self.delete(entry['id'])
        return result

    async def preprocess_parameters(self, params):
        await anonymous(self.cli, params)
        preprocessor = PLUGIN_PREPROCESSORS.get(params.get('name'))
        if preprocessor:
            params = await preprocessor(self.cli, params)
        return params

    async def update(self, id, **params):
        params = await self.preprocess_parameters(params)
        return await super().update(id, **params)


class Plugins(PluginMixin, CrudComponent):

    async def get_for_service(self, plugin_name, service_id):
        plugins = await self.get_list(name=plugin_name, service_id=service_id)
        if not plugins:
            raise KongError('Plugin %s not found' % plugin_name)
        return plugins[0]


class ServicePlugins(PluginMixin, ServiceEntity):

    async def create(self, skip_error=None, **params):
        params['service_id'] = self.root.id
        params = await self.preprocess_parameters(params)
        return await self.execute(
            self.url, 'post', json=params,
            wrap=self.wrap, skip_error=skip_error
        )


class RoutePlugins(PluginMixin, CrudComponent):
    """Plugins associated with a Route
    """
    async def create(self, skip_error=None, **params):
        params['route_id'] = self.root.id
        params = await self.preprocess_parameters(params)
        return await self.execute(
            self.url, 'post', json=params,
            wrap=self.wrap, skip_error=skip_error
        )


class Plugin(KongEntity):

    @classmethod
    def factory(cls, root, data):
        if data['name'] == 'jwt':
            return JWTPlugin(root, data)
        return Plugin(root, data)


class JWTPlugin(Plugin):

    def create_consumer_jwt(self, consumer, **params):
        return self.execute(
            '%s/jwt' % consumer.url, method='POST', json=params
        )

    def get_consumer_by_jwt(self, jwt):
        return self.execute(
            '%s/jwts/%s/consumer' % (self.root.url, jwt), method='GET',
            wrap=self.root.consumers.wrap
        )

    def remove_consumer_jwt(self, consumer, jwt):
        if isinstance(consumer, dict):
            consumer = self.root.consumers.wrap(consumer)
        return self.execute('%s/jwt/%s' % (consumer.url, jwt), method='DELETE')


async def consumer_id_from_username(cli, params):
    if 'consumer_id' in params:
        c = await cli.consumers.get(params['consumer_id'])
        params['consumer_id'] = c['id']
    return params


async def anonymous(cli, params):
    if 'config' in params and 'anonymous' in params['config']:
        c = await cli.consumers.get(params['config']['anonymous'])
        params['config']['anonymous'] = c['id']
    return params


PLUGIN_PREPROCESSORS = {
    'request-termination': consumer_id_from_username
}



import pytest
import asyncio
import aiohttp

from kong.client import Kong, KongError


TESTS = ('test', 'foo', 'pippo')
CONSUMERS = ('an-xxxx-test',)


@pytest.fixture(scope='session')
def loop():
    """Return an instance of the event loop."""
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope='session')
async def cli(loop):
    session = aiohttp.ClientSession(loop=loop)
    async with Kong(session=session) as cli:
        yield cli
        await cleanup(cli)


async def cleanup(cli):
    for name in TESTS:
        try:
            await cli.services.remove(name)
        except KongError as exc:
            if not exc.status == 404:
                raise
    for consumer in CONSUMERS:
        try:
            await cli.consumers.delete(name)
        except KongError as exc:
            if not exc.status == 404:
                raise



import pytest
from kong.plugins import JWTPlugin
from kong.client import KongError


@pytest.fixture(scope='function')
async def clear_consumers(cli):
    consumers = await cli.consumers.get_list()
    for consumer in consumers:
        await cli.consumers.delete(consumer.id)


@pytest.fixture(scope='session')
async def jwt_plugin(cli):
    service = await cli.services.get('test')
    jwt = await service.plugins.create(name='jwt')
    return jwt


async def test_get_for_service(cli, jwt_plugin):
    service = await cli.services.get('test')
    await cli.plugins.get_for_service('jwt', service.id)

    with pytest.raises(KongError):
        await cli.plugins.get_for_service('noplugin', service.id)

    with pytest.raises(KongError):
        await cli.plugins.get_for_service('jwt', 'noservice')


async def test_create_jwt_plugin(jwt_plugin):
    assert isinstance(jwt_plugin, JWTPlugin)


async def test_jwt_create_consumer_jwt(cli, jwt_plugin, clear_consumers):
    consumer = await cli.consumers.create(custom_id='123456')
    token_data = await jwt_plugin.create_consumer_jwt(consumer)
    assert 'key' in token_data
    assert 'secret' in token_data


async def test_jwt_consumer_by_token(cli, jwt_plugin, clear_consumers):
    consumer = await cli.consumers.create(custom_id='123456')
    token = await jwt_plugin.create_consumer_jwt(consumer)
    jwt_consumer = await jwt_plugin.get_consumer_by_jwt(token['key'])
    assert consumer.id == jwt_consumer.id


async def test_jwt_remove_token(cli, jwt_plugin, clear_consumers):
    consumer = await cli.consumers.create(custom_id='123456')
    token = await jwt_plugin.create_consumer_jwt(consumer)
    await jwt_plugin.remove_consumer_jwt(consumer, token['key'])
    with pytest.raises(KongError):
        await jwt_plugin.get_consumer_by_jwt(token)



from .components import CrudComponent, KongEntity, KongError


class Consumers(CrudComponent):

    def wrap(self, data):
        return Consumer(self, data)

    async def apply_json(self, data):
        if not isinstance(data, list):
            data = [data]
        consumers = await self.get_list()
        consumers = dict(((c.username, c) for c in consumers if c.username))
        result = []
        for entry in data:
            if not isinstance(entry, dict):
                raise KongError('dictionary required')
            username = entry.pop('username', None)
            if not username:
                raise KongError('Consumer username is required')
            if username in consumers:
                if entry:
                    consumer = await self.update(username, **entry)
                else:
                    consumer = consumers[username]
                consumers.pop(username)
            else:
                consumer = await self.create(username=username, **entry)

            result.append(consumer.data)

        return result


class Consumer(KongEntity):

    @property
    def username(self):
        return self.data.get('username')




import json
import asyncio

import click

import yaml as _yaml

from kong.client import Kong, KongError


@click.command()
@click.option(
    '--yaml', type=click.File('r'),
    help='Yaml configuration to upload'
)
@click.pass_context
def kong(ctx, yaml):
    return asyncio.get_event_loop().run_until_complete(_run(ctx, yaml))


async def _run(ctx, yaml):
    async with Kong() as cli:
        if yaml:
            try:
                result = await cli.apply_json(_yaml.load(yaml))
                click.echo(json.dumps(result, indent=4))
            except KongError as exc:
                raise click.ClickException(str(exc))
        else:
            click.echo(ctx.get_help())


def main():     # pragma    nocover
    kong()



import os
import pytest
import yaml

from kong.client import KongError


PATH = os.path.dirname(__file__)


def test_client(cli):
    assert cli.session


async def test_create_service(cli):
    c = await cli.services.create(
        name='test', host='example.upstream', port=8080
    )
    assert c.name == 'test'
    assert c.host == 'example.upstream'
    assert c.id


async def test_update_service(cli):
    c = await cli.services.update('test', host='test.upstream')
    assert c.name == 'test'
    assert c.host == 'test.upstream'


async def test_routes(cli):
    c = await cli.services.get('test')
    routes = await c.routes.get_list()
    assert len(routes) == 0
    route = await c.routes.create(hosts=['example.com'])
    assert route['service']['id'] == c.id


async def test_json(cli):
    with open(os.path.join(PATH, 'test.yaml')) as fp:
        manifest = yaml.load(fp)
    await cli.apply_json(manifest)
    srv = await cli.services.get('foo')
    routes = await srv.routes.get_list()
    assert len(routes) == 2
    #
    # check plugins
    plugins = await srv.plugins.get_list()
    assert len(plugins) == 2


async def test_json2(cli):
    with open(os.path.join(PATH, 'test2.yaml')) as fp:
        manifest = yaml.load(fp)
    await cli.apply_json(manifest)
    srv = await cli.services.get('foo')
    routes = await srv.routes.get_list()
    assert len(routes) == 1
    #
    # check plugins
    plugins = await srv.plugins.get_list()
    assert len(plugins) == 2


async def test_add_certificate(cli):
    c = await cli.certificates.create(
        cert='-----BEGIN CERTIFICATE-----...',
        key='-----BEGIN RSA PRIVATE KEY-----...'
    )
    assert c.id
    assert len(c.data['snis']) == 0
    await cli.certificates.delete(c.id)


async def test_hedge_cases(cli):
    with pytest.raises(KongError):
        await cli.apply_json([])

    with pytest.raises(KongError):
        with open(os.path.join(PATH, 'test3.yaml')) as fp:
            await cli.apply_json(yaml.load(fp))

    assert str(cli) == cli.url


async def test_json_plugins(cli):
    with open(os.path.join(PATH, 'test4.yaml')) as fp:
        await cli.apply_json(yaml.load(fp))


async def test_json_route_plugins(cli):
    with open(os.path.join(PATH, 'test6.yaml')) as fp:
        await cli.apply_json(yaml.load(fp))
    with open(os.path.join(PATH, 'test6.yaml')) as fp:
        await cli.apply_json(yaml.load(fp))



import os
import json

import aiohttp

from .components import KongError, KongResponseError
from .services import Services
from .plugins import Plugins
from .consumers import Consumers
from .certificates import Certificates


__all__ = [
    'Kong',
    'KongError',
    'KongResponseError'
]


class Kong:
    url = os.environ.get('KONG_URL', 'http://127.0.0.1:8001')

    def __init__(self, url: str=None, session: object=None) -> None:
        self.url = url or self.url
        self.session = session or aiohttp.ClientSession()
        self.services = Services(self)
        self.plugins = Plugins(self)
        self.consumers = Consumers(self)
        self.certificates = Certificates(self)

    def __repr__(self) -> str:
        return self.url
    __str__ = __repr__

    @property
    def cli(self):
        return self

    async def close(self) -> None:
        await self.session.close()

    async def __aenter__(self) -> object:
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    async def execute(self, url, method=None, headers=None,
                      callback=None, wrap=None, timeout=None, skip_error=None,
                      **kw):
        method = method or 'GET'
        headers = headers or {}
        headers['Accept'] = 'application/json, text/*; q=0.5'
        response = await self.session.request(
            method, url, headers=headers, **kw
        )
        if callback:
            return await callback(response)
        if response.status == 204:
            return True
        data = await response.json()
        if response.status >= 400:
            raise KongResponseError(response, json.dumps(data, indent=4))
        response.raise_for_status()
        return wrap(data) if wrap else data

    async def apply_json(self, config):
        if not isinstance(config, dict):
            raise KongError('Expected a dict got %s' % type(config).__name__)
        result = {}
        for name, data in config.items():
            if not isinstance(data, list):
                data = [data]
            o = getattr(self, name)
            if not o:
                raise KongError('Kong object %s not available' % name)
            result[name] = await o.apply_json(data)
        return result



from click.testing import CliRunner

from kong.cli import kong


def test_empty():
    runner = CliRunner()
    result = runner.invoke(kong, [])
    assert result.exit_code == 0
    assert result.output.startswith('Usage: kong [OPTIONS]')


def test_plugins():
    runner = CliRunner()
    result = runner.invoke(kong, ['--yaml', 'tests/test4.yaml'])
    assert result.exit_code == 0


def test_bad_config():
    runner = CliRunner()
    result = runner.invoke(kong, ['--yaml', 'tests/test5.yaml'])
    assert result.exit_code == 1
    assert result.output.strip() == 'Error: Plugin name not specified'





class KongError(Exception):
    pass


class KongResponseError(KongError):
    def __init__(self, response, message=''):
        self.response = response
        self.message = message

    @property
    def status(self):
        return self.response.status

    def __str__(self):
        return f'{self.response.status}: {self.message}'


class Component:

    def __init__(self, root: object, name: str = None) -> None:
        self.root = root
        self.name = name or self.__class__.__name__.lower()

    def __repr__(self) -> str:
        return self.url

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def cli(self):
        return self.root.cli

    @property
    def url(self) -> str:
        return '%s/%s' % (self.root.url, self.name)

    def execute(self, url: str, method: str=None, **params) -> object:
        return self.root.execute(url, method, **params)

    def apply_json(self, data):
        raise NotImplementedError


class CrudComponent(Component):

    def get_list(self, **params):
        return self.execute(self.url, params=params, wrap=self.wrap_list)

    def get(self, id):
        return self.execute('%s/%s' % (self.url, id), wrap=self.wrap)

    def has(self, id):
        return self.execute('%s/%s' % (self.url, id), 'get',
                            callback=self.head)

    def create(self, skip_error=None, **params):
        return self.execute(self.url, 'post', json=params,
                            wrap=self.wrap, skip_error=skip_error)

    def update(self, id, **params):
        url = '%s/%s' % (self.url, id)
        return self.execute(url, 'patch', json=params, wrap=self.wrap)

    def delete(self, id):
        return self.execute('%s/%s' % (self.url, id), 'delete')

    def wrap_list(self, data):
        return [self.wrap(d) for d in data['data']]

    def wrap(self, data):
        return data

    async def head(self, response):
        if response.status == 404:
            return False
        elif response.status == 200:
            return True
        else:
            raise KongResponseError(response)


class KongEntity:

    def __init__(self, root, data):
        self.root = root
        self.data = data

    def __getitem__(self, item):
        return self.data[item]

    @property
    def cli(self):
        return self.root.cli

    @property
    def id(self):
        return self.data['id']

    @property
    def url(self):
        return '%s/%s' % (self.root.url, self.id)

    def __repr__(self):
        return repr(self.data)

    def execute(self, url, method=None, **params):
        return self.root.execute(url, method, **params)


class ServiceEntity(CrudComponent):

    @property
    def url(self) -> str:
        return '%s/%s' % (self.cli.url, self.name)

    def get_list(self, **params):
        url = '%s/%s' % (self.root.url, self.name)
        return self.execute(url, params=params, wrap=self.wrap_list)

    async def delete_all(self):
        routes = await self.get_list()
        for route in routes:
            await self.delete(route['id'])
        return len(routes)
